/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Wez Furlong <wez@php.net>                                    |
  | Portions influenced by libevent, written by                          |
  |             Niels Provos <provos@citi.umich.edu>                     |
  +----------------------------------------------------------------------+
*/

/* $Id: poll.c,v 1.1 2004/12/13 04:56:31 wez Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_event.h"
#include "php_event_int.h"

#if defined(HAVE_SYS_POLL_H) && defined(HAVE_POLL)

#include <sys/poll.h>

struct poll_state {
	int event_count;
	struct pollfd *fds;
	struct php_event **slot_to_event;
};

static void *poll_event_init(TSRMLS_D)
{
	struct poll_state *st = ecalloc(1, sizeof(*st));
	return st;
}

static void poll_event_fini(void *arg TSRMLS_DC)
{
	struct poll_state *st = (struct poll_state*)arg;
	efree(st->slot_to_event);
	efree(st->fds);
	efree(st);
}

static int poll_event_add(void *arg, struct php_event *ev TSRMLS_DC)
{
	struct poll_state *st = (struct poll_state*)arg;
	return 0;
}

static int poll_event_del(void *arg, struct php_event *ev TSRMLS_DC)
{
	struct poll_state *st = (struct poll_state*)arg;
	return 0;
}

static int poll_event_recalc(void *arg, int max TSRMLS_DC)
{
	struct poll_state *st = (struct poll_state*)arg;
	return 0;
}

static int poll_event_dispatch(void *arg, struct timeval *tv TSRMLS_DC)
{
	struct poll_state *st = (struct poll_state*)arg;
	int res, i, count, msec, nfds;
	struct php_event *ev;
	struct pollfd *pfd;

	if (EVENT_G(event_count) > st->event_count) {
		st->fds = erealloc(st->fds, count * sizeof(struct pollfd));
		st->slot_to_event = erealloc(st->slot_to_event, count * sizeof(struct php_event*));
		st->event_count = count;	
	}
	nfds = 0;
	TAILQ_FOREACH(ev, &EVENT_G(eventqueue), ev_next) {
		pfd = &st->fds[nfds];
			
		pfd->fd = ev->ev_fd;
		pfd->events = 0;

		if (ev->ev_events & EV_WRITE) {
			pfd->events |= POLLOUT;
		}
		if (ev->ev_events & EV_READ) {
			pfd->events |= POLLIN;
		}
		if (ev->ev_events & EV_EXCEP) {
			pfd->events |= POLLPRI;
		}

		if (pfd->events) {
			st->slot_to_event[nfds] = ev;
			pfd->revents = 0;
			nfds++;
		}
	}

	msec = tv ? (tv->tv_sec * 1000) + (tv->tv_usec / 1000) : -1;
	res = poll(st->fds, nfds, msec);	

	if (res == -1) {
		if (errno != EINTR) {
			return -1;
		}
		return 0;
	}

	if (res == 0) {
		return 0;
	}

	for (i = 0; i < nfds; i++) {
		pfd = &st->fds[nfds];

		res = 0;
		if (pfd->revents & (POLLERR|POLLHUP)) {
			res |= EV_READ|EV_WRITE;
		}
		if (pfd->revents & POLLIN) {
			res |= EV_READ;
		}
		if (pfd->revents & POLLOUT) {
			res |= EV_WRITE;
		}

		ev = st->slot_to_event[i];
		res &= ev->ev_events;

		if (res == 0) {
			continue;
		}

		event_active(ev, res, 1 TSRMLS_CC);			
	}
	return 0;	
}

const struct php_eventop php_event_poll_engine = {
	"poll",
	poll_event_init,
	poll_event_fini,
	poll_event_add,
	poll_event_del,
	poll_event_recalc,
	poll_event_dispatch
};

#endif

